using System;
using System.Collections.Generic;
using System.Threading.Tasks;
using System.Data.SqlClient;
using System.Linq;
using System.Data;

using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.CardDav;
using ITHit.WebDAV.Server.Acl;
using ITHit.WebDAV.Server.Class1;


namespace CalDAVServer.SqlStorage.CardDav
{
    /// <summary>
    /// Represents a CardDAV address book (address book folder).
    /// Instances of this class correspond to the following path: [DAVLocation]/addressbooks/[AddressbookFolderId]
    /// </summary>
    public class AddressbookFolder : DavHierarchyItem, IAddressbookFolderAsync, ICurrentUserPrincipalAsync
    {
        /// <summary>
        /// Loads address book folder by ID. Returns null if addressbook folder was not found.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/> class.</param>
        /// <param name="addressbookFolderId">ID of the address book folder to load.</param>
        /// <returns><see cref="IAddressbookFolderAsync"/> instance.</returns>
        public static async Task<IAddressbookFolderAsync> LoadByIdAsync(DavContext context, Guid addressbookFolderId)
        {
            string sql = string.Format("SELECT * FROM [card_AddressbookFolder] WHERE [AddressbookFolderId]=@AddressbookFolderId");

            return (await LoadAsync(context, sql, "@AddressbookFolderId", addressbookFolderId)).FirstOrDefault();
        }

        /// <summary>
        /// Loads all address books.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/> class.</param>
        /// <returns>List of <see cref="IAddressbookFolderAsync"/> items.</returns>
        public static async Task<IEnumerable<IAddressbookFolderAsync>> LoadAllAsync(DavContext context)
        {
            return await LoadAsync(context, "SELECT * FROM [card_AddressbookFolder]");
        }

        /// <summary>
        /// Loads address book by SQL.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/> class.</param>
        /// <param name="sql">SQL that queries [card_AddressbookFolder] table.</param>
        /// <param name="prms">List of SQL parameters.</param>
        /// <returns>List of <see cref="IAddressbookFolderAsync"/> items.</returns>
        private static async Task<IEnumerable<IAddressbookFolderAsync>> LoadAsync(DavContext context, string sql, params object[] prms)
        {
            IList<IAddressbookFolderAsync> addressbookFolders = new List<IAddressbookFolderAsync>();

            using (SqlDataReader reader = await context.ExecuteReaderAsync(sql, prms))
            {
                DataTable addressbooks = new DataTable();
                addressbooks.Load(reader);

                foreach (DataRow rowAddressbookFolder in addressbooks.Rows)
                {
                    Guid addressbookFolderId = rowAddressbookFolder.Field<Guid>("AddressbookFolderId");
                    addressbookFolders.Add(new AddressbookFolder(context, addressbookFolderId, rowAddressbookFolder));
                }
            }

            return addressbookFolders;
        }

        /// <summary>
        /// Creates new address book folder.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/> class.</param> 
        /// <param name="name">Addressbook folder name.</param>
        public static async Task CreateAddressbookFolderAsync(DavContext context, string name)
        {
            await context.ExecuteNonQueryAsync("INSERT INTO [card_AddressbookFolder] ([AddressbookFolderId], [Name], [Description]) VALUES (@AddressbookFolderId, @Name, @Description )",
                "@AddressbookFolderId", Guid.NewGuid(), "@Name", name, "@Description", "");
        }

        /// <summary>
        /// This address book folder ID.
        /// </summary>
        private readonly Guid addressbookFolderId;

        /// <summary>
        /// Contains data from [card_AddressbookFolder] table.
        /// </summary>
        private readonly DataRow rowAddressbookFolder;

        /// <summary>
        /// Gets display name of the address book.
        /// </summary>
        /// <remarks>CalDAV clients typically never request this property.</remarks>
        public override string Name
        {
            get { return rowAddressbookFolder != null ? rowAddressbookFolder.Field<string>("Name") : null; }
        }

        /// <summary>
        /// Gets item path.
        /// </summary>
        public override string Path
        {
            get
            {
                return string.Format("{0}{1}/", AddressbooksRootFolder.AddressbooksRootFolderPath, addressbookFolderId);
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="AddressbookFolder"/> class from database source.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/> class.</param>
        /// <param name="addressbookFolderId">Address book folder ID.</param>
        /// <param name="addressbook">Contains data from [card_AddressbookFolder] table. Used when listing folder content.</param>
        private AddressbookFolder(DavContext context, Guid addressbookFolderId, DataRow addressbook)
            : base(context)
        {
            this.addressbookFolderId = addressbookFolderId;
            this.rowAddressbookFolder = addressbook;
        }

        /// <summary>
        /// Returns a list of address book files that correspont to the specified list of item paths.
        /// </summary>
        /// <remarks>
        /// <para>
        /// This method is called by the Engine during <b>addressbook-multiget</b> call.
        /// </para>
        /// <para>
        /// For each item from the <b>pathList</b> parameter return an item that corresponds to path or <b>null</b> if the item is not found.
        /// </para>
        /// </remarks>
        /// <param name="pathList">Addressbook files path list.</param>
        /// <param name="propNames">
        /// Properties requested by the client. You can use this as a hint about what properties will be called by 
        /// the Engine for each item that are returned from this method.
        /// </param>
        /// <returns>List of address book files. Returns <b>null</b> for any item that is not found.</returns>
        public async Task<IEnumerable<ICardFileAsync>> MultiGetAsync(IEnumerable<string> pathList, IEnumerable<PropertyName> propNames)
        {
            // Get list of UIDs from path list.
            IEnumerable<string> uids = pathList.Select(a => System.IO.Path.GetFileNameWithoutExtension(a));

            return await CardFile.LoadByUidsAsync(Context, uids, PropsToLoad.All);
        }

        /// <summary>
        /// Returns a list of address book files that match specified filter. 
        /// </summary>
        /// <remarks>
        /// <para>
        /// This method is called by the Engine during <b>addressbook-query</b> call.
        /// </para>
        /// </remarks>
        /// <param name="rawQuery">
        /// Raw query sent by the client.
        /// </param>
        /// <param name="propNames">
        /// Properties requested by the client. You can use this as a hint about what properties will be called by 
        /// the Engine for each item that are returned from this method.
        /// </param>
        /// <returns>List of address book files. Returns <b>null</b> for any item that is not found.</returns>
        public async Task<IEnumerable<ICardFileAsync>> QueryAsync(string rawQuery, IEnumerable<PropertyName> propNames)
        {
            // For the sake of simplicity we just call GetChildren returning all items. 
            // Typically you will return only items that match the query.
            return (await GetChildrenAsync(propNames.ToList())).Cast<ICardFileAsync>();
        }

        /// <summary>
        /// Provides a human-readable description of the address book collection.
        /// </summary>
        public string AddressbookDescription 
        {
            get { return rowAddressbookFolder.Field<string>("Description"); }
        }

        /// <summary>
        /// Retrieves children of this folder.
        /// </summary>
        /// <param name="propNames">List of properties to retrieve with the children. They will be queried by the engine later.</param>
        /// <returns>Children of the folder.</returns>
        public async Task<IEnumerable<IHierarchyItemAsync>> GetChildrenAsync(IList<PropertyName> propNames)
        {
            // Here we enumerate all events and to-dos contained in this address book.
            // You can filter children items in this implementation and 
            // return only items that you want to be available for this 
            // particular user.

            // Typically only getcontenttype and getetag properties are requested in GetChildren call by CalDAV/CardDAV clients.
            // The iCalendar/vCard (calendar-data/address-data) is typically requested not in GetChildren, but in a separate multiget 
            // report, in MultiGetAsync() method call, that follows this request.

            // Bynari submits PROPFIND without props - Engine will request getcontentlength

            IList<IHierarchyItemAsync> children = new List<IHierarchyItemAsync>();
            return await CardFile.LoadByAddressbookFolderIdAsync(Context, addressbookFolderId, PropsToLoad.Minimum);
        }

        /// <summary>
        /// Creates a file that contains event or to-do item in this address book.
        /// </summary>
        /// <param name="name">Name of the file. This is a vCard UID ending with '.vcf'.</param>
        /// <returns>The newly created file.</returns>
        /// <remarks></remarks>
        public async Task<IFileAsync> CreateFileAsync(string name)
        {
            // The actual event or to-do object is created in datatbase in CardFile.Write call.
            return CardFile.CreateCardFile(Context, addressbookFolderId);
        }

        /// <summary>
        /// Creating new folders is not allowed in address book folders.
        /// </summary>
        /// <param name="name">Name of the folder.</param>
        public async Task CreateFolderAsync(string name)
        {
            throw new DavException("Not allowed.", DavStatus.NOT_ALLOWED);
        }

        /// <summary>
        /// Move this folder to folder <paramref name="destFolder"/>.
        /// </summary>
        /// <param name="destFolder">Destination folder.</param>
        /// <param name="destName">Name for this folder at destination.</param>
        /// <param name="multistatus">Instance of <see cref="MultistatusException"/>
        /// to fill with errors ocurred while moving child items.</param>
        /// <returns></returns>
        public override async Task MoveToAsync(IItemCollectionAsync destFolder, string destName, MultistatusException multistatus)
        {
            // Here we support only addressbooks renaming.
            await Context.ExecuteNonQueryAsync("UPDATE [card_AddressbookFolder] SET Name=@Name WHERE AddressbookFolderId=@AddressbookFolderId", "@Name", destName, "@AddressbookFolderId", addressbookFolderId);
        }

        /// <summary>
        /// Deletes this address book.
        /// </summary>
        /// <param name="multistatus"><see cref="MultistatusException"/> to populate with child files and folders failed to delete.</param>
        public override async Task DeleteAsync(MultistatusException multistatus)
        {
            await Context.ExecuteNonQueryAsync("DELETE FROM [card_AddressbookFolder] WHERE [AddressbookFolderId]=@AddressbookFolderId", "@AddressbookFolderId", addressbookFolderId);
        }

        /// <summary>
        /// Gets values of all properties or selected properties for this item.
        /// </summary>
        /// <param name="names">Property names which values are
        /// requested by WebDAV client. If a property does not exist for this hierarchy item
        /// then the property value shall not be returned.
        /// </param>
        /// <param name="allprop">If it is true, besides properties listed in props you need to return
        /// all properties you think may be useful to client.
        /// </param>
        /// <returns></returns>
        public override async Task<IEnumerable<PropertyValue>> GetPropertiesAsync(IList<PropertyName> names, bool allprop)
        {
            IList<PropertyValue> propVals = await GetPropertyValuesAsync(
                    "SELECT [Name], [Namespace], [PropVal] FROM [card_AddressbookFolderProperty] WHERE [AddressbookFolderId] = @AddressbookFolderId",
                    "@AddressbookFolderId", addressbookFolderId);

            if (allprop)
            {
                return propVals;
            }
            else
            {
                IList<PropertyValue> requestedPropVals = new List<PropertyValue>();
                foreach (PropertyValue p in propVals)
                {
                    if (names.Contains(p.QualifiedName))
                    {
                        requestedPropVals.Add(p);
                    }
                }
                return requestedPropVals;
            }
        }

        /// <summary>
        /// Adds, modifies and removes properties for this item.
        /// </summary>
        /// <param name="setProps">List of properties to be set.</param>
        /// <param name="delProps">List of property names to be removed. Properties that don't exist shall be skipped.</param>
        /// <param name="multistatus">Information about errors.</param>
        public override async Task UpdatePropertiesAsync(
            IList<PropertyValue> setProps,
            IList<PropertyName> delProps,
            MultistatusException multistatus)
        {
            foreach (PropertyValue p in setProps)
            {
                await SetPropertyAsync(p); // create or update property
            }

            foreach (PropertyName p in delProps)
            {
                await RemovePropertyAsync(p.Name, p.Namespace);
            }
        }

        /// <summary>
        /// Reads <see cref="PropertyValue"/> from database by executing SQL command.
        /// </summary>
        /// <param name="command">Command text.</param>
        /// <param name="prms">SQL parameter pairs: SQL parameter name, SQL parameter value</param>
        /// <returns>List of <see cref="PropertyValue"/>.</returns>
        private async Task<IList<PropertyValue>> GetPropertyValuesAsync(string command, params object[] prms)
        {
            List<PropertyValue> l = new List<PropertyValue>();
            
            using (SqlDataReader reader = await Context.ExecuteReaderAsync(command, prms))
            {
                while (reader.Read())
                {
                    string name = reader.GetString(reader.GetOrdinal("Name"));
                    string ns = reader.GetString(reader.GetOrdinal("Namespace"));
                    string value = reader.GetString(reader.GetOrdinal("PropVal"));
                    l.Add(new PropertyValue(new PropertyName(name, ns), value));
                }
            }

            return l;
        }

        private async Task SetPropertyAsync(PropertyValue prop)
        {
            string selectCommand =
                @"SELECT Count(*) FROM [card_AddressbookFolderProperty]
                  WHERE [AddressbookFolderId] = @AddressbookFolderId AND [Name] = @Name AND [Namespace] = @Namespace";

            int count = await Context.ExecuteScalarAsync<int>(
                selectCommand,
                "@AddressbookFolderId"  , addressbookFolderId,
                "@Name"                 , prop.QualifiedName.Name,
                "@Namespace"            , prop.QualifiedName.Namespace);

            // insert
            if (count == 0)
            {
                string insertCommand = @"INSERT INTO [card_AddressbookFolderProperty] ([AddressbookFolderId], [Name], [Namespace], [PropVal])
                                          VALUES(@AddressbookFolderId, @Name, @Namespace, @PropVal)";

                await Context.ExecuteNonQueryAsync(
                    insertCommand,
                    "@PropVal"              , prop.Value,
                    "@AddressbookFolderId"  , addressbookFolderId,
                    "@Name"                 , prop.QualifiedName.Name,
                    "@Namespace"            , prop.QualifiedName.Namespace);
            }
            else
            {
                // update
                string command = @"UPDATE [card_AddressbookFolderProperty]
                      SET [PropVal] = @PropVal
                      WHERE [AddressbookFolderId] = @AddressbookFolderId AND [Name] = @Name AND [Namespace] = @Namespace";

                await Context.ExecuteNonQueryAsync(
                    command,
                    "@PropVal"              , prop.Value,
                    "@AddressbookFolderId"  , addressbookFolderId,
                    "@Name"                 , prop.QualifiedName.Name,
                    "@Namespace"            , prop.QualifiedName.Namespace);
            }
        }

        private async Task RemovePropertyAsync(string name, string ns)
        {
            string command = @"DELETE FROM [card_AddressbookFolderProperty]
                              WHERE [AddressbookFolderId] = @AddressbookFolderId
                              AND [Name] = @Name
                              AND [Namespace] = @Namespace";

            await Context.ExecuteNonQueryAsync(
                command,
                "@AddressbookFolderId"  , addressbookFolderId,
                "@Name"                 , name,
                "@Namespace"            , ns);
        }
    }
}
